import logging
from pathlib import Path

import yaml

from truss.base.truss_config import TrussConfig
from truss.templates.control.control.helpers.custom_types import (
    Action,
    ConfigPatch,
    ModelCodePatch,
    PackagePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_handle.patch.truss_dir_patch_applier import TrussDirPatchApplier

TEST_LOGGER = logging.getLogger("test_logger")


def test_model_code_patch(custom_model_truss_dir: Path):
    applier = TrussDirPatchApplier(custom_model_truss_dir, TEST_LOGGER)
    applier(
        [
            Patch(
                type=PatchType.MODEL_CODE,
                body=ModelCodePatch(
                    action=Action.UPDATE, path="model.py", content="test_content"
                ),
            )
        ]
    )
    assert (custom_model_truss_dir / "model" / "model.py").read_text() == "test_content"


def test_packages_patch(custom_model_truss_dir: Path):
    applier = TrussDirPatchApplier(custom_model_truss_dir, TEST_LOGGER)
    applier(
        [
            Patch(
                type=PatchType.PACKAGE,
                body=PackagePatch(
                    action=Action.UPDATE, path="user_package.py", content="import sys"
                ),
            )
        ]
    )
    assert (
        custom_model_truss_dir / "packages" / "user_package.py"
    ).read_text() == "import sys"


def test_python_requirement_patch(custom_model_truss_dir: Path):
    req = "git+https://github.com/huggingface/transformers.git"
    applier = TrussDirPatchApplier(custom_model_truss_dir, TEST_LOGGER)
    config = yaml.safe_load((custom_model_truss_dir / "config.yaml").read_text())
    config["requirements"] = [req]
    applier(
        [
            Patch(
                type=PatchType.PYTHON_REQUIREMENT,
                body=PythonRequirementPatch(action=Action.ADD, requirement=req),
            ),
            Patch(
                type=PatchType.CONFIG,
                body=ConfigPatch(
                    action=Action.UPDATE, config=config, path="config.yaml"
                ),
            ),
        ]
    )
    assert TrussConfig.from_yaml(
        custom_model_truss_dir / "config.yaml"
    ).requirements == [req]


def test_system_requirement_patch(custom_model_truss_dir: Path):
    applier = TrussDirPatchApplier(custom_model_truss_dir, TEST_LOGGER)
    config = yaml.safe_load((custom_model_truss_dir / "config.yaml").read_text())
    config["system_packages"] = ["curl"]
    applier(
        [
            Patch(
                type=PatchType.CONFIG,
                body=ConfigPatch(
                    action=Action.UPDATE, config=config, path="config.yaml"
                ),
            ),
            Patch(
                type=PatchType.SYSTEM_PACKAGE,
                body=SystemPackagePatch(action=Action.ADD, package="curl"),
            ),
        ]
    )
    assert TrussConfig.from_yaml(
        custom_model_truss_dir / "config.yaml"
    ).system_packages == ["curl"]



import sys
from typing import Optional, Tuple, cast

import rich_click as click
from rich.console import Console

from truss.remote.baseten.remote import BasetenRemote

POLL_INTERVAL_SEC = 2


def check_is_interactive() -> bool:
    """Detects if CLI is operated interactively by human, so we can ask things,
    that we would want to skip for automated subprocess/CI contexts."""
    return sys.stdin.isatty() and sys.stdout.isatty()


def get_most_recent_job(
    console: Console,
    remote_provider: BasetenRemote,
    project_id: Optional[str],
    job_id: Optional[str],
) -> Tuple[str, str]:
    if not project_id or not job_id:
        jobs = remote_provider.api.search_training_jobs(
            project_id=project_id, job_id=job_id
        )
        if not jobs:
            raise click.UsageError("No jobs found.")
        if len(jobs) > 1:
            sorted_jobs = sorted(jobs, key=lambda x: x["created_at"], reverse=True)
            job = sorted_jobs[0]
            console.print(
                f"Multiple jobs found. Showing the most recently created job: {job['id']}",
                style="yellow",
            )
        else:
            job = jobs[0]
        project_id = cast(str, job["training_project"]["id"])
        job_id = cast(str, job["id"])
    return project_id, job_id



from typing import Dict, List, Optional, Union

import pydantic

from truss.base import custom_types, truss_config

DEFAULT_LORA_RANK = 16


class SecretReference(custom_types.SafeModel):
    name: str


class Compute(custom_types.SafeModel):
    node_count: int = 1
    cpu_count: int = 1
    memory: str = "2Gi"
    accelerator: Optional[truss_config.AcceleratorSpec] = None

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        if self.accelerator and self.accelerator.accelerator:
            data["accelerator"] = {
                "accelerator": self.accelerator.accelerator.value,
                "count": self.accelerator.count,
            }
        return data

    def to_truss_config(self) -> truss_config.Resources:
        if self.accelerator:
            return truss_config.Resources(
                cpu=str(self.cpu_count),
                memory=self.memory,
                accelerator=self.accelerator,
                node_count=self.node_count,
            )
        return truss_config.Resources(
            cpu=str(self.cpu_count), memory=self.memory, node_count=self.node_count
        )


class CheckpointingConfig(custom_types.SafeModel):
    enabled: bool = False
    checkpoint_path: Optional[str] = None


class Runtime(custom_types.SafeModel):
    start_commands: List[str] = []
    environment_variables: Dict[str, Union[str, SecretReference]] = {}
    enable_cache: bool = False
    checkpointing_config: CheckpointingConfig = CheckpointingConfig()


class Image(custom_types.SafeModel):
    base_image: str


class TrainingJob(custom_types.SafeModel):
    image: Image
    compute: Compute = Compute()
    runtime: Runtime = Runtime()

    def model_dump(self, *args, **kwargs):
        data = super().model_dump(*args, **kwargs)
        data["compute"] = self.compute.model_dump()
        return data


class TrainingProject(custom_types.SafeModel):
    name: str
    # TrainingProject is the wrapper around project config and job config. However, we exclude job
    # in serialization so just TrainingProject metadata is included in API requests.
    job: TrainingJob = pydantic.Field(exclude=True)


class Checkpoint(custom_types.SafeModel):
    training_job_id: str
    id: str
    name: str
    lora_rank: Optional[int] = (
        None  # lora rank will be fetched through the API if available.
    )

    def to_truss_config(self) -> truss_config.Checkpoint:
        return truss_config.Checkpoint(
            id=f"{self.training_job_id}/{self.id}", name=self.id
        )


class CheckpointList(custom_types.SafeModel):
    download_folder: str = truss_config.DEFAULT_TRAINING_CHECKPOINT_FOLDER
    base_model_id: Optional[str] = None
    checkpoints: List[Checkpoint] = []

    def to_truss_config(self) -> truss_config.CheckpointList:
        checkpoints = [checkpoint.to_truss_config() for checkpoint in self.checkpoints]
        return truss_config.CheckpointList(
            checkpoints=checkpoints, download_folder=self.download_folder
        )


class DeployCheckpointsRuntime(custom_types.SafeModel):
    environment_variables: Dict[str, Union[str, SecretReference]] = {}


class DeployCheckpointsConfig(custom_types.SafeModel):
    checkpoint_details: Optional[CheckpointList] = None
    model_name: Optional[str] = None
    deployment_name: Optional[str] = None
    runtime: Optional[DeployCheckpointsRuntime] = None
    compute: Optional[Compute] = None
